/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#pragma once
#include "duskgl.h"
#include <archive.h>
#include <archive_entry.h>

#define ASSET_BUFFER_SIZE 32768

typedef struct {
  FILE *file;
  struct archive *archive;
  struct archive_entry *entry;
  uint8_t buffer[ASSET_BUFFER_SIZE];

  // ?
  char_t path[FILENAME_MAX];
} asset_t;

extern asset_t ASSET;

/**
 * Initializes the asset manager.
 */
void assetInit();

/**
 * Opens an asset by its filename (within the asset archive). Asset paths should
 * always use the unix forward slash '/' as a path separator.
 * 
 * @param path The path to the asset within the archive.
 */
void assetOpen(const char *path);

/**
 * Returns the size of the asset.
 * 
 * @return The size of the asset.
 */
size_t assetGetSize();

/**
 * Reads the asset into the buffer.
 * 
 * @param buffer The buffer to read the asset into.
 * @param bufferSize The size of the buffer.
 * @return The amount of data read.
 */
size_t assetRead(uint8_t *buffer, const size_t bufferSize);

/**
 * Reads ahead in the buffer until either the end of the buffer, or the
 * specified character is found. Return value will be -1 if the character was
 * not found.
 * 
 * Buffer can be NULL if you just want to skip ahead.
 * 
 * Returned value will be either the amount of data read into the buffer, which
 * excludes the extra 1 character that was read from the asset. If the character
 * was not found, -1 will be returned.
 * 
 * @param buffer Buffer to read into.
 * @param c Character to read until.
 * @param maxLength Maximum length to read.
 * @return -1 if the character was not found, otherwise the amount of data read.
 */
size_t assetReadUntil(uint8_t *buffer, const char c, const size_t maxLength);

/**
 * Skips ahead in the buffer by the specified length.
 * 
 * @param length The length to skip ahead by.
 */
void assetSkip(const size_t length);

/**
 * Closes the asset.
 */
void assetClose();

/**
 * Destroys and cleans up the asset manager.
 */
void assetDispose();

/**
 * Internal read method provided to libarchive api.
 * 
 * @param archive The archive to read from.
 * @param data The data to read into.
 * @param buffer The buffer to read from.
 * @return The amount of data read.
 */
ssize_t assetArchiveRead(
  struct archive *archive,
  void *data,
  const void **buffer
);

/**
 * Internal seek method provided to libarchive api.
 * 
 * @param archive The archive to seek in.
 * @param data The data to seek in.
 * @param offset Offset bytes to seek.
 * @param whence Relative to whence to seek.
 * @return The new position.
 */
int64_t assetArchiveSeek(
  struct archive *archive,
  void *data,
  int64_t offset,
  int32_t whence
);

/**
 * Internal open method provided to libarchive api.
 * 
 * @param archive The archive to open.
 * @param data The data to open.
 * @return The result of the open.
 */
int32_t assetArchiveOpen(struct archive *a, void *data);

/**
 * Internal close method provided to libarchive api.
 * 
 * @param archive The archive to close.
 * @param data The data to close.
 * @return The result of the close.
 */
int32_t assetArchiveClose(struct archive *a, void *data);