/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "assert/assert.h"
#include "display/window.h"
#include "display/render.h"
#include "game.h"
#include "asset.h"
#include "console/console.h"

char_t EXECUTABLE_PATH[FILENAME_MAX];
char_t EXECUTABLE_DIRECTORY[FILENAME_MAX];

int32_t main(int32_t argc, char_t **argv) {
  // Get launch args
  assertTrue(argc > 0, "argc is not valid!");
  assertNotNull(argv, "argv is not valid!");
  assertNotNull(argv[0], "argv[0] is not valid!");
  assertStrLen(argv[0], FILENAME_MAX, "argv[0] is too long!");
  assertStrLenMin(argv[0], 1, "argv[0] is empty!");

  // Store the executable path and directory.
  strcpy(EXECUTABLE_PATH, argv[0]);
  strcpy(EXECUTABLE_DIRECTORY, dirname(EXECUTABLE_PATH));

  // Init systems
  int32_t ret;
  if((ret = windowInit()) != 0) {
    return -1;
  }
  gameInit();

  // Print OS Information
  const char* glfwVersionString = glfwGetVersionString();
  consolePrint("GLFW: %s", glfwVersionString);

  // Init asset and render systems.
  assetInit();
  renderInit();

  // Prepare for time tracking
  double_t time, newTime;
  float_t fDelta;
  int32_t updateResult;

  // Main loop
  while(!windowShouldClose()) {
    // Determine the delta.
    newTime = glfwGetTime();
    fDelta = (float_t)(newTime - time);
    time = newTime;

    gameUpdate(fDelta);

    // Draw
    renderUpdate();
    windowUpdate();
  }

  renderDispose();
  windowDispose();
  assetDispose();
  gameDispose();

  return 0;
}